package net.sf.sido.io;

import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;

import net.sf.sido.DataFactory;
import net.sf.sido.DataObject;
import net.sf.sido.DataType;
import net.sf.sido.Property;
import net.sf.sido.PropertyMetaInfos;
import net.sf.sido.array.Array;
import net.sf.sido.editor.DataEditor;
import net.sf.sido.support.DOCache;

/**
 * Default implementation for a {@link DataObjectWriter} that relies on writing some attributes and
 * elements to the output, in a format-independent way.
 * 
 * @author Damien Coraboeuf
 * @param <T>
 *            Type of element to write into (XML element, JSON node, etc.)
 * @param <N>
 *            Type that identifies the name of an element
 */
public abstract class AbstractDataObjectWriter<T, N> implements DataObjectWriter<T, N> {

	protected abstract void add(T e, N propertyName, T propertyValue);

	protected abstract void createArray(T e, N pName, List<T> elements);

	protected T createElement(DOReferenceManagement referenceManagement, DOCache cache, N name, DataObject o, boolean typeRequired) {
		// Creates the element
		T e = createElement(name);
		// Type
		DataType type = o.getSidoType();
		if (typeRequired) {
			setDOType(e, type);
		}
		// The given cache belongs to the upper level of the graph
		// Cached object
		int id = o.getSidoId();
		if (cache.contains(type, id)) {
			setDOReference(e, id);
			// OK
			return e;
		}
		// ID
		setDOId(e, id);
		// Puts the current object in the cache in all cases
		cache.put(id, o);
		// For each property
		for (Property<?> property : type.getProperties()) {
			// Cache to use
			DOCache propertyCache;
			if (referenceManagement == DOReferenceManagement.GLOBAL) {
				// Keeps the same cache over and over
				propertyCache = cache;
			} else {
				// Gets a copy of this cache
				propertyCache = new DOCache(cache);
			}
			// Qualified name for the property
			N pName = getPropertyName(e, property);
			// Gets the value for this property
			Object value = o.getValue(property);
			if (value != null) {
				boolean doProperty = property.getMetaInfo().getBoolean(PropertyMetaInfos.DO, false);
				// Array
				if (property.isArray()) {
					Array<?> array = (Array<?>) value;
					List<T> elements = new ArrayList<T>();
					// For each element
					for (Object element : array) {
						// Data object
						if (doProperty) {
							// Type for the property
							DataType propertyType = property.getMetaInfo().getObject(PropertyMetaInfos.ELEMENT_TYPE, null);
							boolean anonymous = (propertyType == null);
							boolean abstractType = (propertyType != null && propertyType.isAbstractType());
							// Child name
							N cName = getDOArrayName();
							// Child element
							T eChild = createElement(referenceManagement, propertyCache, cName, (DataObject) element, anonymous || abstractType);
							// Appends the child
							elements.add(eChild);
						}
						// Simple property
						else {
							// Child name
							N cName = getSimpleArrayName();
							// Gets the value as a string
							T eChild = createValue(o, cName, element);
							// Adds to the array
							elements.add(eChild);
						}
					}
					// Creates the array
					createArray(e, pName, elements);
				}
				// Data object
				else if (doProperty) {
					DataObject child = (DataObject) value;
					// Type for the property
					DataType propertyType = property.getMetaInfo().getObject(PropertyMetaInfos.DO_TYPE, null);
					boolean anonymous = (propertyType == null);
					boolean abstractType = (propertyType != null && propertyType.isAbstractType());
					// Creates the node
					T eChild = createElement(referenceManagement, propertyCache, pName, child, anonymous || abstractType);
					// Appends the child
					add(e, pName, eChild);
				}
				// Simple value
				else {
					T eChild = createValue(o, pName, value);
					if (eChild != null) {
						// Appends the child
						add(e, pName, eChild);
					}
				}
			}
		}
		// OK
		return e;
	}

	protected abstract T createElement(N name);

	@Override
	public T createElement(N name, DataObject o) {
		return createElement(name, o, DOReferenceManagement.STACK);
	}

	@Override
	public T createElement(N name, DataObject o, DOReferenceManagement referenceManagement) {
		return createElement(referenceManagement, new DOCache(), name, o, true);
	}

	protected abstract T createValue(DataObject o, N pName, Object value);

	protected abstract N getDOArrayName();

	protected abstract N getPropertyName(T e, Property<?> property);

	protected abstract N getSimpleArrayName();

	protected <V> String getText(DataFactory factory, V value) {
		// Type
		@SuppressWarnings("unchecked")
		Class<V> type = (Class<V>) value.getClass();
		// Gets the editor
		DataEditor<V> editor = factory.getEditor(type);
		// String value
		String string = editor.toString(value);
		// OK
		return string;
	}

	protected abstract void setDOId(T e, int id);

	protected abstract void setDOReference(T e, int id);

	protected abstract void setDOType(T e, DataType type);

	@Override
	public void write(OutputStream out, DataObject o) throws IOException {
		write(out, o, DOReferenceManagement.STACK);
	}

}
