package net.sf.sido.support;

import java.io.IOException;
import java.io.ObjectStreamException;
import java.io.Serializable;
import java.util.*;

import net.sf.sido.*;
import net.sf.sido.array.Array;

import net.sf.sido.io.SiDODeserializerContext;
import org.apache.commons.lang.ObjectUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;

import com.google.common.collect.ImmutableMap;

public class DefaultDataObject extends AbstractDataObject implements Map<String, Object>, Serializable {

    private static final long serialVersionUID = 1;

	public static final String PROPERTY_SEPARATOR = ".";

	private final boolean immutable;
	private final Map<String, Object> values;

	private ParentLink parentLink;

    // Deserialization fields
    private transient boolean t_immutable;
    private transient Map<String, Object> t_values;

	protected DefaultDataObject(DataObject caller, DataType type, Map<String, Object> values, boolean immutable) {
		super(type.newInstanceId(), type);
		Map<String, Object> thisValues = new HashMap<String, Object>();
		for(Map.Entry<String, Object> entry : values.entrySet()) {
			String name = entry.getKey();
			Object value = entry.getValue();
			if (value instanceof DataObject) {
				DataObject d = (DataObject) value;
				// Old parent
				ParentLink oldLink = d.getParentLink();
				if (immutable) {
					d = d.immutable();
				} else {
					d = d.copy();
				}
				// Parent
				if (oldLink != null && oldLink.getParent() == caller) {
					d.setParentLink(oldLink.changeParent(this));
				}
				// OK
				value = d;
			}
			thisValues.put(name, value);
		}
		if (immutable) {
			this.values = ImmutableMap.copyOf(thisValues);
		} else {
			this.values = thisValues;
		}
		this.immutable = immutable;
	}

	public DefaultDataObject(int id, DataType type) {
		super(id, type);
		values = new HashMap<String, Object>();
		immutable = false;
	}

    private void writeObject(java.io.ObjectOutputStream out) throws IOException {
        out.writeBoolean(immutable);
        out.writeObject(values);
    }

    private void readObject(java.io.ObjectInputStream in)
     throws IOException, ClassNotFoundException {
        t_immutable = in.readBoolean();
        t_values = (Map<String, Object>) in.readObject();
    }

    private Object readResolve() throws ObjectStreamException {
        DataFactory t_dataFactory = SiDODeserializerContext.getDataFactory();
        DataType t_type = t_dataFactory.getType(t_typename);
        if (t_type == null) {
            throw new DOTypeNotFoundException(t_typename);
        }
        DataObject o = t_type.newInstance();
        for (Map.Entry<String, Object> entry : t_values.entrySet()) {
            String name = entry.getKey();
            Object value = entry.getValue();
            o.setObject(name, value);
        }
        if (t_immutable) {
            o = o.immutable();
        }
        t_typename = null;
        t_values = null;
        return o;
    }

	@Override
	public Object clone() throws CloneNotSupportedException {
		return copy();
	}

	@Override
	public DataObject copy() {
		return new DefaultDataObject(this, getSidoType(), values, false);
	}

	@Override
	public void init(boolean deep) {
		Collection<Property<?>> properties = getSidoType().getProperties();
		for(Property<?> property : properties) {
			init(property, deep);
		}
	}

	protected <T> void init(Property<T> property, boolean deep) {
		// Gets any existing value
		T value = getValue(property);
		// If not filled and mandatory
		if (value == null && !property.isNullable()) {
			// Array
			if (property.isArray()) {
				Array<T> array = Array.empty();
				values.put(property.getName(), array);
			} else {
				// Gets a default value
				value = property.getDefaultValue(deep);
				// Sets the value
				if (value != null) {
					setValue(property, value);
				}
			}
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T> T getObject(String name) {
		Validate.notEmpty(name, "Property name must not be blank");
		if (name.contains(PROPERTY_SEPARATOR)) {
			String prefix = StringUtils.substringBefore(name, PROPERTY_SEPARATOR);
			String leftOver = StringUtils.substringAfter(name, PROPERTY_SEPARATOR);
			// Gets the dataobject
			DataObject child = getDataObject(prefix);
			// If not defined, returns null
			if (child == null) {
				return null;
			} else {
				// Calls the child
				return (T) child.getObject(leftOver);
			}
		} else {
			T value = (T) values.get(name);
			value = getFactory().onPropertyGet(this, name, value);
			return value;
		}
	}

	@Override
	public ParentLink getParentLink() {
		return parentLink;
	}

	@Override
	public DataObject immutable() {
		if (immutable) {
			return this;
		} else {
			return new DefaultDataObject(this, getSidoType(), values, true);
		}
	}

	@Override
	public boolean isImmutable() {
		return immutable;
	}

	@Override
	public <T> DataObject setObject(String name, T value) {
		if (immutable) {
			throw new DOImmutableException();
		} else {
			Validate.notEmpty(name, "Property name must not be blank");
			if (name.contains(PROPERTY_SEPARATOR)) {
				String prefix = StringUtils.substringBefore(name, PROPERTY_SEPARATOR);
				String leftOver = StringUtils.substringAfter(name, PROPERTY_SEPARATOR);
				// Gets the dataobject
				DataObject child = getDataObject(prefix);
				// If not defined, does nothing
				if (child != null) {
					// Calls the child
					child.setObject(leftOver, value);
				}
			} else {
				DataFactory factory = getFactory();
				// Old value
				@SuppressWarnings("unchecked")
				T oldValue = (T) values.get(name);
				// Pre-control
				T actualValue = factory.onPropertySetPre(this, name, oldValue, value);
				// Sets the value
				if (actualValue != null) {
					values.put(name, actualValue);
				} else {
					values.remove(name);
				}
				// Post-control
				factory.onPropertySetPost(this, name, oldValue, actualValue);
			}
			// Returns itself for chaining
			return this;
		}
	}

	@Override
	public void setParentLink(ParentLink parentLink) {
		this.parentLink = parentLink;
	}

	@Override
	public ValidationResults validate() {
		ValidationDataObjectVisitor visitor = new ValidationDataObjectVisitor();
		visit(visitor);
		return visitor.getResults();
	}

	@Override
	public void visit(DataObjectVisitor visitor) {
		visit(visitor, new DefaultVisitorContext());
	}

	@Override
	public void visit(DataObjectVisitor visitor, VisitorContext context) {
		// Ok for this node
		if (!visitor.getCache().contains(this)) {
			// Visited
			visitor.getCache().put(this);
			// Visits this node first
			visitor.visit(context, this);
			// Gets the type
			DataType type = getSidoType();
			// For all properties
			Collection<Property<?>> properties = type.getProperties();
			for(Property<?> property : properties) {
				visit(context, property, visitor);
			}
		}
	}

	protected <T> void visit(VisitorContext context, Property<T> property, DataObjectVisitor visitor) {
		String name = property.getName();
		// Value for this property
		T value = getValue(property);
		// Visit the property
		visitor.visit(context, this, property);
		// Validate DO property
		if (value != null && value instanceof DataObject) {
			DataObject child = (DataObject) value;
			// Recursive visit
			child.visit(visitor, new DefaultVisitorContext(context, name));
		}
		// Validate DO array
		else if (value != null && property.isArray() && property.getMetaInfo().getBoolean(PropertyMetaInfos.DO, false)) {
			@SuppressWarnings("unchecked")
			Array<DataObject> array = (Array<DataObject>) value;
			int index = 0;
			for(DataObject child : array) {
				String newPath = name + "[" + index + "]";
				child.visit(visitor, new DefaultVisitorContext(context, newPath));
				index++;
			}
		}
	}

	@Override
	public int size() {
		return values.size();
	}

	@Override
	public boolean isEmpty() {
		return values.isEmpty();
	}

	@Override
	public boolean containsKey(Object key) {
		return values.containsKey(key);
	}

	@Override
	public boolean containsValue(Object value) {
		return values.containsValue(value);
	}

	/**
	 * Only {@link String} keys are accepted.
	 */
	@Override
	public Object get(Object key) {
		if (key instanceof String) {
			return get((String) key);
		} else {
			throw new IllegalArgumentException("DefaultDataObject accept only String as key type");
		}
	}

	@Override
	public Object put(String key, Object value) {
		Object previous = get(key);
		setValue(key, value);
		return previous;
	}

	@Override
	public Object remove(Object key) {
		if (key instanceof String) {
			return setValue((String) key, null);
		} else {
			throw new IllegalArgumentException("DefaultDataObject accept only String as key type");
		}
	}

	@Override
	public void putAll(Map<? extends String, ? extends Object> m) {
		Set<? extends Map.Entry<? extends String, ? extends Object>> entries = m.entrySet();
		for(Entry<? extends String, ? extends Object> entry : entries) {
			String key = entry.getKey();
			Object value = entry.getValue();
			setValue(key, value);
		}
	}

	@Override
	public void clear() {
		// Gets a copy of all keys
		Set<String> keys = new HashSet<String>(values.keySet());
		for(String key : keys) {
			setValue(key, null);
		}
	}

	@Override
	public Set<String> keySet() {
		return values.keySet();
	}

	@Override
	public Collection<Object> values() {
		return values.values();
	}

	@Override
	public Set<java.util.Map.Entry<String, Object>> entrySet() {
		return values.entrySet();
	}

    @Override
    public Map<String, Difference> diff(DataObject that) {
        Map<String, Difference> diffs = new HashMap<String, Difference>();
        // Gets all properties using raw values
        Map<String, Object> thisValues = this.values;
        Map<String, Object> thatValues = (Map<String,Object>) that;
        // Merge of key values
        Set<String> allKeys = new HashSet<String>(thisValues.keySet());
        allKeys.addAll(thatValues.keySet());
        // For all keys
        for (String key : allKeys) {
            // Gets both values
            Object value1 = thisValues.get(key);
            Object value2 = thatValues.get(key);
            // At least one value is != null
            if (value1 != null || value2 != null) {
                // If both values are != null
                if (value1 != null && value2 != null) {
                    // Needs further comparison
                    Collection<Difference> subdiff = diff (key, value1, value2);
                    if (subdiff != null) {
                        for (Difference diff: subdiff) {
                            diffs.put(diff.getProperty(), diff);
                        }
                    }
                }
                // There is a difference
                else {
                    diffs.put(key, new Difference(key, value1, value2));
                }
            }
        }
        // OK
        return diffs;
    }

    protected Collection<Difference> diff (String key, Object value1, Object value2) {
        if (value1 instanceof DataObject) {
            Map<String, Difference> diff = ((DataObject) value1).diff((DataObject) value2);
            Collection<Difference> diffs = new ArrayList<Difference>();
            for (Map.Entry<String, Difference> d : diff.entrySet()) {
                diffs.add(d.getValue().prefix(key + "."));
            }
            return diffs;
        } else if (value1 instanceof Array) {
            return diffArray(key, (Array) value1, (Array) value2);
        } else {
            // Simple type
            if (!ObjectUtils.equals(value1, value2)) {
                return Collections.singleton(new Difference(key, value1, value2));
            } else {
                return null;
            }
        }
    }

    protected <T> Collection<Difference> diffArray (String key, Array<T> values1, Array<T> values2) {
        Collection<Difference> diffs = new ArrayList<Difference>();
        int len1 = values1.getLength();
        int len2 = values2.getLength();
        int maxlen = Math.max(len1, len2);
        for (int i = 0 ; i < maxlen ; i++) {
            T value1 = values1.getOrNull(i);
            T value2 = values2.getOrNull(i);
            if (value1 != null || value2 != null) {
                String arrayKey = key + "[" + i + "]";
                if (value1 != null && value2 != null) {
                    Collection<Difference> diff = diff(arrayKey, value1, value2);
                    if (diff != null) {
                        diffs.addAll(diff);
                    }
                } else {
                    diffs.add(new Difference(arrayKey, value1, value2));
                }
            }
        }
        return diffs;
    }
}
