package net.sf.sido.support;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;

import net.sf.sido.DOTypeNotExtensibleException;
import net.sf.sido.DataObject;
import net.sf.sido.DataSchema;
import net.sf.sido.DataType;
import net.sf.sido.Property;
import net.sf.sido.UndefinedProperty;
import net.sf.sido.array.Array;
import net.sf.sido.DOCircularInheritanceException;

import org.apache.commons.lang.ObjectUtils;
import org.apache.commons.lang.StringUtils;

import com.google.common.base.Predicate;

public class DefaultDataType implements DataType {

	private final DataSchema schema;
	private final String name;

	private final AtomicInteger id = new AtomicInteger();

	private Map<String, Property<?>> properties = new LinkedHashMap<String, Property<?>>();
	private boolean allowExtensions;
	private boolean closed;
	private boolean abstractType;
	private DataType parentType;

	public DefaultDataType(DataSchema schema, String name) {
		this.schema = schema;
		this.name = name;
	}

	public <T> void addProperty(UndefinedProperty<T> undefinedProperty) {
		if (closed) {
			throw new DOClosedTypeException(name);
		}
		// Defines the property holder
		undefinedProperty.setHolderType(this);
		// Indexes the properties
		properties.put(undefinedProperty.getName(), undefinedProperty);
	}

	@Override
	public void close() {
		closed = true;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof DataType) {
			DataType that = (DataType) obj;
			return StringUtils.equals(name, that.getName()) && ObjectUtils.equals(schema, that.getSchema());
		} else {
			return false;
		}
	}

	protected <T> DataType extend(DefaultProperty<T> property) {
		if (isExtensible()) {
			property.setHolderType(this);
			properties.put(property.getName(), property);
			return this;
		} else {
			throw new DOTypeNotExtensibleException(name);
		}
	}

	@Override
	public <T> DataType extend(String name, Class<T> type, boolean nullable, Predicate<T> constraint) {
		DefaultProperty<T> property = new DefaultProperty<T>(name, nullable, type, true, constraint,
				new SimplePropertyController<T>(type));
		return extend(property);
	}

	@Override
	public DataType extend(String name, DataType type, boolean nullable, boolean reverse,
			Predicate<DataObject> constraint) {
		DefaultProperty<DataObject> property = new DefaultProperty<DataObject>(name, nullable, DataObject.class, true,
				constraint, new DOPropertyController(type, reverse));
		return extend(property);
	}

	@Override
	public <T> DataType extendArray(String name, Class<T> type, boolean nullable, Predicate<Array<T>> constraint) {
		DefaultProperty<Array<T>> property = new DefaultProperty<Array<T>>(name, nullable, Array.class, true,
				constraint, new ArrayPropertyController<T>(type));
		return extend(property);
	}

	@Override
	public DataType extendArray(String name, DataType type, boolean nullable, Predicate<Array<DataObject>> constraint) {
		DefaultProperty<Array<DataObject>> property = new DefaultProperty<Array<DataObject>>(name, nullable,
				Array.class, true, constraint, new ArrayDOPropertyController(type));
		return extend(property);
	}

	@Override
	public String getName() {
		return name;
	}

	@Override
	public DataType getParentType() {
		return parentType;
	}

	@Override
	public Collection<Property<?>> getProperties() {
		Collection<Property<?>> collection;
		if (parentType != null) {
			collection = parentType.getProperties();
		} else {
			collection = new LinkedHashSet<Property<?>>();
		}
		collection.addAll(properties.values());
		return collection;
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T> Property<T> getProperty(String name) {
		Property<T> p = (Property<T>) properties.get(name);
		if (p == null && parentType != null) {
			return parentType.getProperty(name);
		} else {
			return p;
		}
	}

	@Override
	public DataSchema getSchema() {
		return schema;
	}

	@Override
	public List<DataType> getSubTypes() {
		return getSchema().getFactory().getSubTypes(this);
	}

	@Override
	public int hashCode() {
		return name.hashCode();
	}

	@Override
	public boolean isAbstractType() {
		return abstractType;
	}

	@Override
	public boolean isExtensible() {
		return allowExtensions;
	}

	@Override
	public boolean isSubClass(DataType type) {
		// Same type
		if (this.equals(type)) {
			return true;
		} else if (parentType != null) {
			return parentType.isSubClass(type);
		} else {
			return false;
		}
	}

	@Override
	public DataObject newInstance() {
		return new DefaultDataObject(newInstanceId(), this);
	}

	@Override
	public int newInstanceId() {
		return id.incrementAndGet();
	}

	public void setAbstractType(boolean abstractType) {
		this.abstractType = abstractType;
	}

	public void setAllowExtensions(boolean allowExtensions) {
		this.allowExtensions = allowExtensions;
	}

	public void setParentType(DataType parentType) {
		// Checks for double update
		if (this.parentType != null) {
			throw new IllegalStateException("Parent type has already been set");
		}
		// Checks for circularity
		if (parentType != null) {
			DataType type = parentType;
			while (type != null) {
				if (type == this) {
					throw new DOCircularInheritanceException(getQualifiedName());
				} else {
					type = type.getParentType();
				}
			}
		}
		// OK
		this.parentType = parentType;
	}

	@Override
	public String toString() {
        return getQualifiedName();
	}

    @Override
    public String getQualifiedName() {
		return String.format("%s%s%s", schema.getName(), DataSchema.QUALIFIER_SEPARATOR, name);
    }

    @Override
	public DataObject newInstance(boolean init) {
		return newInstance(init, false);
	}

	@Override
	public DataObject newInstance(boolean init, boolean deep) {
		DataObject o = newInstance();
		if (init) {
			o.init(deep);
		}
		return o;
	}

}
