package net.sf.sido.support;

import java.math.BigDecimal;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import net.sf.sido.DataObject;
import net.sf.sido.DataSchema;
import net.sf.sido.DataType;
import net.sf.sido.DataTypeBuilder;
import net.sf.sido.PropertyAdapter;
import net.sf.sido.UndefinedProperty;
import net.sf.sido.array.Array;

import org.joda.time.DateTime;
import org.joda.time.Duration;
import org.joda.time.Period;

import com.google.common.base.Predicate;

public class DefaultDataTypeBuilder implements DataTypeBuilder {

	private final String name;
	private final DefaultDataType type;

	private final Map<String, PropertyAdapter<?>> adapters;

	public DefaultDataTypeBuilder(DataSchema schema, String name) {
		this.name = name;
		type = new DefaultDataType(schema, name);
		// Default adapters
		adapters = new HashMap<String, PropertyAdapter<?>>();
		adapters.put("Byte", new NumericPropertyAdapter<Byte>("byteValue"));
		adapters.put("Short", new NumericPropertyAdapter<Short>("shortValue"));
		adapters.put("Integer", new NumericPropertyAdapter<Integer>("intValue"));
		adapters.put("Long", new NumericPropertyAdapter<Long>("longValue"));
		adapters.put("Float", new NumericPropertyAdapter<Float>("floatValue"));
		adapters.put("Double", new NumericPropertyAdapter<Double>("doubleValue"));
		adapters.put("BigDecimal", new DecimalPropertyAdapter());
	}

	@Override
	public String getName() {
		return name;
	}

	@Override
	public DataTypeBuilder abstractType() {
		type.setAbstractType(true);
		return this;
	}

	@Override
	public DataTypeBuilder addAnonymousArrayProperty(String name, boolean nullable, Predicate<Array<DataObject>> constraint) {
		addProperty(new DefaultProperty<Array<DataObject>>(name, nullable, Array.class, false, constraint, new AnonymousArrayPropertyController()));
		return this;
	}

	@Override
	public DataTypeBuilder addAnonymousProperty(String name, boolean reverse, boolean nullable, Predicate<DataObject> constraint) {
		// Creates the property
		DefaultProperty<DataObject> property = new DefaultProperty<DataObject>(name, nullable, DataObject.class, false, constraint, new AnonymousPropertyController(reverse));
		// OK
		return addProperty(property);
	}

	@Override
	public <T> DataTypeBuilder addArrayProperty(String name, Class<T> elementType) {
		return addArrayProperty(name, elementType, false);
	}

	@Override
	public <T> DataTypeBuilder addArrayProperty(String name, Class<T> elementType, boolean nullable) {
		return addArrayProperty(name, elementType, nullable, null);
	}

	@Override
	public <T> DataTypeBuilder addArrayProperty(String name, Class<T> elementType, boolean nullable, Predicate<Array<T>> constraint) {
		addProperty(new DefaultProperty<Array<T>>(name, nullable, Array.class, false, constraint, new ArrayPropertyController<T>(elementType)));
		return this;
	}

	@Override
	public <T> DataTypeBuilder addArrayProperty(String name, DataType elementType) {
		return addArrayProperty(name, elementType, false);
	}

	@Override
	public DataTypeBuilder addArrayProperty(String name, DataType elementType, boolean nullable) {
		return addArrayProperty(name, elementType, nullable, null);
	}

	@Override
	public DataTypeBuilder addArrayProperty(String name, DataType elementType, boolean nullable, Predicate<Array<DataObject>> constraint) {
		addProperty(new DefaultProperty<Array<DataObject>>(name, nullable, Array.class, false, constraint, new ArrayDOPropertyController(elementType)));
		return this;
	}

	@Override
	public DataTypeBuilder addBooleanProperty(String name) {
		return addBooleanProperty(name, false);
	}

	@Override
	public DataTypeBuilder addBooleanProperty(String name, boolean nullable) {
		return addBooleanProperty(name, nullable, null);
	}

	@Override
	public DataTypeBuilder addBooleanProperty(String name, boolean nullable, Predicate<Boolean> constraint) {
		return addProperty(name, nullable, Boolean.class, constraint);
	}

	@Override
	public DataTypeBuilder addByteProperty(String name) {
		return addByteProperty(name, false);
	}

	@Override
	public DataTypeBuilder addByteProperty(String name, boolean nullable) {
		return addByteProperty(name, nullable, null);
	}

	@Override
	public DataTypeBuilder addByteProperty(String name, boolean nullable, Predicate<Byte> constraint) {
		return addProperty(name, nullable, Byte.class, constraint);
	}

	@Override
	public DataTypeBuilder addDateTimeProperty(String name) {
		return addDateTimeProperty(name, false);
	}

	@Override
	public DataTypeBuilder addDateTimeProperty(String name, boolean nullable) {
		return addDateTimeProperty(name, nullable, null);
	}

	@Override
	public DataTypeBuilder addDateTimeProperty(String name, boolean nullable, Predicate<DateTime> constraint) {
		return addProperty(name, nullable, DateTime.class, constraint);
	}

	@Override
	public DataTypeBuilder addDecimalProperty(String name) {
		return addDecimalProperty(name, false);
	}

	@Override
	public DataTypeBuilder addDecimalProperty(String name, boolean nullable) {
		return addDecimalProperty(name, nullable, null);
	}

	@Override
	public DataTypeBuilder addDecimalProperty(String name, boolean nullable, Predicate<BigDecimal> constraint) {
		return addProperty(name, nullable, BigDecimal.class, constraint);
	}

	@Override
	public DataTypeBuilder addDOProperty(String name, DataType type) {
		return addDOProperty(name, type, true);
	}

	@Override
	public DataTypeBuilder addDOProperty(String name, DataType type, boolean reverse) {
		return addDOProperty(name, type, reverse, false);
	}

	@Override
	public DataTypeBuilder addDOProperty(String name, DataType type, boolean reverse, boolean nullable) {
		return addDOProperty(name, type, reverse, nullable, null);
	}

	@Override
	public DataTypeBuilder addDOProperty(String name, DataType type, boolean reverse, boolean nullable, Predicate<DataObject> constraint) {
		// Creates the property
		DefaultProperty<DataObject> property = new DefaultProperty<DataObject>(name, nullable, DataObject.class, false, constraint, new DOPropertyController(type, reverse));
		// OK
		return addProperty(property);
	}

	@Override
	public DataTypeBuilder addDoubleProperty(String name) {
		return addDoubleProperty(name, false);
	}

	@Override
	public DataTypeBuilder addDoubleProperty(String name, boolean nullable) {
		return addDoubleProperty(name, nullable, null);
	}

	@Override
	public DataTypeBuilder addDoubleProperty(String name, boolean nullable, Predicate<Double> constraint) {
		return addProperty(name, nullable, Double.class, constraint);
	}

	@Override
	public DataTypeBuilder addDurationProperty(String name) {
		return addDurationProperty(name, false);
	}

	@Override
	public DataTypeBuilder addDurationProperty(String name, boolean nullable) {
		return addDurationProperty(name, nullable, null);
	}

	@Override
	public DataTypeBuilder addDurationProperty(String name, boolean nullable, Predicate<Duration> constraint) {
		return addProperty(name, nullable, Duration.class, constraint);
	}

	@Override
	public DataTypeBuilder addFloatProperty(String name) {
		return addFloatProperty(name, false);
	}

	@Override
	public DataTypeBuilder addFloatProperty(String name, boolean nullable) {
		return addFloatProperty(name, nullable, null);
	}

	@Override
	public DataTypeBuilder addFloatProperty(String name, boolean nullable, Predicate<Float> constraint) {
		return addProperty(name, nullable, Float.class, constraint);
	}

	@Override
	public DataTypeBuilder addIntegerProperty(String name) {
		return addIntegerProperty(name, false);
	}

	@Override
	public DataTypeBuilder addIntegerProperty(String name, boolean nullable) {
		return addIntegerProperty(name, nullable, null);
	}

	@Override
	public DataTypeBuilder addIntegerProperty(String name, boolean nullable, Predicate<Integer> constraint) {
		return addProperty(name, nullable, Integer.class, constraint);
	}

	@Override
	public DataTypeBuilder addLongProperty(String name) {
		return addLongProperty(name, false);
	}

	@Override
	public DataTypeBuilder addLongProperty(String name, boolean nullable) {
		return addLongProperty(name, nullable, null);
	}

	@Override
	public DataTypeBuilder addLongProperty(String name, boolean nullable, Predicate<Long> constraint) {
		return addProperty(name, nullable, Long.class, constraint);
	}

	@Override
	public DataTypeBuilder addPeriodProperty(String name) {
		return addPeriodProperty(name, false);
	}

	@Override
	public DataTypeBuilder addPeriodProperty(String name, boolean nullable) {
		return addPeriodProperty(name, nullable, null);
	}

	@Override
	public DataTypeBuilder addPeriodProperty(String name, boolean nullable, Predicate<Period> constraint) {
		return addProperty(name, nullable, Period.class, constraint);
	}

	@Override
	public <T> DataTypeBuilder addProperty(String name, boolean nullable, Class<T> type, Predicate<T> constraint) {
		// Adapter?
		PropertyAdapter<T> adapter = getPropertyAdapter(type);
		// Creates the property
		UndefinedProperty<T> property = new DefaultProperty<T>(name, nullable, type, false, constraint, Collections.singleton(new SimplePropertyController<T>(type)), adapter);
		// OK
		return addProperty(property);
	}

	@SuppressWarnings("unchecked")
	protected <T> PropertyAdapter<T> getPropertyAdapter(Class<T> type) {
		return (PropertyAdapter<T>) adapters.get(type.getSimpleName());
	}

	@Override
	public <T> DataTypeBuilder addProperty(UndefinedProperty<T> property) {
		String name = property.getName();
		if (type.getProperty(name) != null) {
			throw new IllegalStateException(String.format("Type %s contains already the %s property", type, name));
		} else {
			type.addProperty(property);
			return this;
		}
	}

	@Override
	public DataTypeBuilder addShortProperty(String name) {
		return addShortProperty(name, false);
	}

	@Override
	public DataTypeBuilder addShortProperty(String name, boolean nullable) {
		return addShortProperty(name, nullable, null);
	}

	@Override
	public DataTypeBuilder addShortProperty(String name, boolean nullable, Predicate<Short> constraint) {
		return addProperty(name, nullable, Short.class, constraint);
	}

	@Override
	public DataTypeBuilder addStringProperty(String name) {
		return addStringProperty(name, false);
	}

	@Override
	public DataTypeBuilder addStringProperty(String name, boolean nullable) {
		return addStringProperty(name, nullable, null);
	}

	@Override
	public DataTypeBuilder addStringProperty(String name, boolean nullable, Predicate<String> constraint) {
		return addProperty(name, nullable, String.class, constraint);
	}

	@Override
	public DataTypeBuilder allowExtensions() {
		type.setAllowExtensions(true);
		return this;
	}

	@Override
	public DataType getType() {
		return type;
	}

	@Override
	public DataTypeBuilder setParentType(DataType parentType) {
		type.setParentType(parentType);
		return this;
	}
}
